﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Text;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Windows.Forms;

namespace NameCard {
    enum GISelectionStatus {
        NoItemFirstOfAll,
        None,
        OneSetOfCharacters,
        OneBitmapImage,
        OneGroupedItem,
        MultipleItems
    }

    class Editor : IDisposable {
        private const string CLASS_NAME = "Editor";
        private const float MM_INCH = 25.4f;

        private bool _disposed = false;
        private FileInfo _file = null;
        private NameCard _card = null;
        private List<GraphicItem> _selecteditems = new List<GraphicItem>();
        private FormMain _drawsurface = null;
        private Stack<byte[]> _undolist = new Stack<byte[]>();
        private IFormatter _formatter = new BinaryFormatter();

        public const string FileFilter = "名刺データ (*.nmc)|*.nmc|すべてのファイル (*.*)|*.*";

        public Editor(FormMain drawsurface) {
            _drawsurface = drawsurface;
            _card = new NameCard();
            _drawsurface.Invalidate();
        }
        public Editor(FormMain drawsurface, FileInfo file) {
            _drawsurface = drawsurface;
            using(var s = file.OpenRead())
                _card = (NameCard)_formatter.Deserialize(s);
            _drawsurface.Invalidate();
            _file = file;
        }
        protected void Dispose(bool disposing) {
            if(disposing) {
                //Free manage resource
                _card.Dispose();
            }
            //Free unmanage resource
            _disposed = true;
        }
        #region IDisposable Member
        public void Dispose() {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion
        ~Editor() {
            Dispose(false);
        }

        private void ClearUndoList() {
            _undolist.Clear();
            _drawsurface.EnableUndo();
            GC.Collect();
        }
        private void PushToUndoList() {
            using(var s = new MemoryStream()) {
                var package = new UndoPackage(_card, _selecteditems);
                _formatter.Serialize(s, package);
                _undolist.Push(s.ToArray());
            }
            _drawsurface.EnableUndo();
        }
        public void PopFromUndoList() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            using(Stream s = new MemoryStream(_undolist.Pop())) {
                var package = (UndoPackage)_formatter.Deserialize(s);
                package.Restore(ref _card, ref _selecteditems);
            }
            _drawsurface.EnableUndo();
            _drawsurface.Invalidate();
        }
        public int UndoListCount {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _undolist.Count;
            }
        }

        public void Save() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            SaveAs(_file);
            ClearUndoList();
        }
        public void SaveAs(FileInfo file) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            using(Stream s = file.OpenWrite()) _formatter.Serialize(s, _card);
            _file = file;
            ClearUndoList();
        }

        public GISelectionStatus Hit(PointF posmm, bool ctrlkey) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            GraphicItem gi = _card.ItemHit(posmm);
            if(ctrlkey) {
                if(gi != null) {
                    if(_selecteditems.IndexOf(gi) >= 0)
                        _selecteditems.Remove(gi);
                    else
                        _selecteditems.Add(gi);
                }
            } else {
                _selecteditems.Clear();
                if(gi != null && _selecteditems.IndexOf(gi) < 0)
                    _selecteditems.Add(gi);
            }
            _drawsurface.Invalidate();
            return SelectionStatus;
        }
        public GISelectionStatus SelectAll() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus == GISelectionStatus.NoItemFirstOfAll)
                throw new Exception("There is no item to be selected");
            _selecteditems.Clear();
            foreach(var gi in _card) _selecteditems.Add(gi);
            _drawsurface.Invalidate();
            return SelectionStatus;
        }
        public GISelectionStatus SelectionStatus {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                if(_card.Count == 0)
                    return GISelectionStatus.NoItemFirstOfAll;
                if(_selecteditems.Count == 0)
                    return GISelectionStatus.None;
                if(_selecteditems.Count > 1)
                    return GISelectionStatus.MultipleItems;
                if(_selecteditems[0] is Characters)
                    return GISelectionStatus.OneSetOfCharacters;
                if(_selecteditems[0] is BitmapImage)
                    return GISelectionStatus.OneBitmapImage;
                else
                    return GISelectionStatus.OneGroupedItem;
            }
        }

        public GISelectionStatus InsertCharacters(string text, FontInfo font, PointF posmm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            var sizemm = GetCharactersSize(_drawsurface, font, text);
            PushToUndoList();
            _card.Add(new Characters(text, font, posmm, sizemm));
            _drawsurface.Invalidate();
            return SelectionStatus;
        }
        private SizeF GetCharactersSize(Control ds, FontInfo fi, string text) {
            using(var f = fi.CreateFont())
            using(var sf = StringFormat.GenericTypographic)
            using(var g = ds.CreateGraphics()) {
                g.PageUnit = GraphicsUnit.Millimeter;
                g.TextRenderingHint = TextRenderingHint.AntiAlias;
                return g.MeasureString(text, f, new PointF(0, 0), sf);
            }
        }
        public GISelectionStatus InsertImage(Bitmap image, SizeF sizemm, PointF posmm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            PushToUndoList();
            _card.Add(new BitmapImage(image, posmm, sizemm));
            _drawsurface.Invalidate();
            return SelectionStatus;
        }
        public GISelectionStatus Remove() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus == GISelectionStatus.NoItemFirstOfAll ||
               SelectionStatus == GISelectionStatus.None)
                throw new Exception("There is no item to be removed");
            PushToUndoList();
            foreach(var gi in _selecteditems) _card.Remove(gi);
            _selecteditems.Clear();
            _drawsurface.Invalidate();
            return SelectionStatus;
        }

        public void GetSelectedCharacters(out string text, out FontInfo font, out PointF posmm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(GISelectionStatus.OneSetOfCharacters != SelectionStatus)
                throw new Exception("Only one set of characters is not selected.");
            text = ((Characters)_selecteditems[0]).Text;
            font = ((Characters)_selecteditems[0]).Font;
            posmm = ((Characters)_selecteditems[0]).Positionmm;
        }
        public void EditCharacters(string text, FontInfo font, PointF posmm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(GISelectionStatus.OneSetOfCharacters != SelectionStatus)
                throw new Exception("Only one set of characters is not selected.");
            var sizemm = GetCharactersSize(_drawsurface, font, text);
            PushToUndoList();
            if(((Characters)_selecteditems[0]).SetContents(text, font, posmm, sizemm))
                _drawsurface.Invalidate();
            else
                PopFromUndoList();
        }
        public void GetSelectedImage(out Bitmap image, out SizeF sizemm, out PointF posmm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(GISelectionStatus.OneBitmapImage != SelectionStatus)
                throw new Exception("Only one bitmap image is not selected.");
            var gi = (BitmapImage)_selecteditems[0];
            image = gi.Image;
            sizemm = gi.Sizemm;
            posmm = gi.Positionmm;
        }
        public void EditImage(SizeF sizemm, PointF posmm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(GISelectionStatus.OneBitmapImage != SelectionStatus)
                throw new Exception("Only one bitmap image is not selected.");
            var gi = (BitmapImage)_selecteditems[0];
            PushToUndoList();
            if(gi.SetSize(sizemm) | gi.SetPosition(posmm))
                _drawsurface.Invalidate();
            else
                PopFromUndoList();
        }

        public GISelectionStatus CopyPaste() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(GISelectionStatus.OneSetOfCharacters == SelectionStatus) goto Accept;
            if(GISelectionStatus.OneBitmapImage == SelectionStatus) goto Accept;
            if(GISelectionStatus.OneGroupedItem == SelectionStatus) goto Accept;
            throw new Exception("Only one item is not selected.");

        Accept:
            PushToUndoList();
            using(var s = new MemoryStream()) {
                _formatter.Serialize(s, _selecteditems[0]);
                _selecteditems.Clear();
                s.Seek(0, SeekOrigin.Begin);
                var gi = (GraphicItem)_formatter.Deserialize(s);
                gi.SetPosition(new PointF(0f, 0f));
                _card.Add(gi);
                _selecteditems.Add(gi);
            }
            _drawsurface.Invalidate();
            return SelectionStatus;
        }
        public GISelectionStatus Group() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus != GISelectionStatus.MultipleItems)
                throw new Exception("There are not items to be grouped.");
            PushToUndoList();
            var gi = new GroupedItem(_selecteditems);
            _selecteditems.Clear();
            _selecteditems.Add(gi);
            foreach(var subgi in gi) _card.Remove(subgi);
            _card.Add(gi);
            _drawsurface.Invalidate();
            return SelectionStatus;
        }
        public GISelectionStatus Dissolve() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(GISelectionStatus.OneGroupedItem != SelectionStatus)
                throw new Exception("One GroupedItem is not selected.");
            PushToUndoList();
            var gi = (GroupedItem)_selecteditems[0];
            _selecteditems.Clear();
            foreach(var subgi in gi) {
                _selecteditems.Add(subgi);
                _card.Add(subgi);
            }
            _card.Remove(gi);
            gi.Clear();
            _drawsurface.Invalidate();
            return SelectionStatus;
        }

        public enum Direction {
            Left,
            Center,
            Right,
            Up,
            Middle,
            Down
        }
        public void Move(Direction dir, float lenmm, bool repeat) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus == GISelectionStatus.NoItemFirstOfAll ||
               SelectionStatus == GISelectionStatus.None)
                throw new Exception("There is no item to be moved");
            if(!repeat) PushToUndoList();
            foreach(var gi in _selecteditems) {
                var posmm = gi.Positionmm;
                switch(dir) {
                    case Direction.Left:
                        posmm.X -= lenmm; break;
                    case Direction.Right:
                        posmm.X += lenmm; break;
                    case Direction.Up:
                        posmm.Y -= lenmm; break;
                    case Direction.Down:
                        posmm.Y += lenmm; break;
                }
                gi.SetPosition(posmm);
            }
            _drawsurface.Invalidate();
        }
        public void Align(Direction dir) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus == GISelectionStatus.NoItemFirstOfAll ||
               SelectionStatus == GISelectionStatus.None)
                throw new Exception("There are not items to be aligned.");
            var changed = false;
            PushToUndoList();
            switch(dir) {
                case Direction.Left:
                    changed = AlignLeft(); break;
                case Direction.Center:
                    changed = AlignCenter(); break;
                case Direction.Right:
                    changed = AlignRight(); break;
                case Direction.Up:
                    changed = AlignUp(); break;
                case Direction.Middle:
                    changed = AlignMiddle(); break;
                case Direction.Down:
                    changed = AlignDown(); break;
            }
            if(changed) _drawsurface.Invalidate();
            else PopFromUndoList();
        }
        private bool AlignLeft() {
            var changed = false;
            if(SelectionStatus == GISelectionStatus.MultipleItems) {
                var b = float.PositiveInfinity;
                foreach(var gi in _selecteditems) if(gi.Leftmm < b) b = gi.Leftmm;
                foreach(var gi in _selecteditems) {
                    var posmm = gi.Positionmm;
                    posmm.X = b;
                    changed |= gi.SetPosition(posmm);
                }
            } else {
                var posmm = _selecteditems[0].Positionmm;
                posmm.X = 0f;
                changed = _selecteditems[0].SetPosition(posmm);
            }
            return changed;
        }
        private bool AlignCenter() {
            var changed = false;
            if(SelectionStatus == GISelectionStatus.MultipleItems) {
                var b = 0f;
                foreach(var gi in _selecteditems) b += gi.Leftmm + gi.Widthmm / 2;
                b /= _selecteditems.Count;
                foreach(var gi in _selecteditems) {
                    var posmm = gi.Positionmm;
                    posmm.X = b - gi.Widthmm / 2;
                    changed |= gi.SetPosition(posmm);
                }
            } else {
                var posmm = _selecteditems[0].Positionmm;
                posmm.X = _card.Sizemm.Width / 2 - _selecteditems[0].Widthmm / 2;
                changed = _selecteditems[0].SetPosition(posmm);
            }
            return changed;
        }
        private bool AlignRight() {
            var changed = false;
            if(SelectionStatus == GISelectionStatus.MultipleItems) {
                var b = float.NegativeInfinity;
                foreach(var gi in _selecteditems) {
                    var right = gi.Leftmm + gi.Widthmm;
                    if(b < right) b = right;
                }
                foreach(var gi in _selecteditems) {
                    var posmm = gi.Positionmm;
                    posmm.X = b - gi.Widthmm;
                    changed |= gi.SetPosition(posmm);
                }
            } else {
                var posmm = _selecteditems[0].Positionmm;
                posmm.X = _card.Sizemm.Width - _selecteditems[0].Widthmm;
                changed = _selecteditems[0].SetPosition(posmm);
            }
            return changed;
        }
        private bool AlignUp() {
            var changed = false;
            if(SelectionStatus == GISelectionStatus.MultipleItems) {
                var b = float.PositiveInfinity;
                foreach(var gi in _selecteditems) if(gi.Topmm < b) b = gi.Topmm;
                foreach(var gi in _selecteditems) {
                    var posmm = gi.Positionmm;
                    posmm.Y = b;
                    changed |= gi.SetPosition(posmm);
                }
            } else {
                var posmm = _selecteditems[0].Positionmm;
                posmm.Y = 0f;
                changed = _selecteditems[0].SetPosition(posmm);
            }
            return changed;
        }
        private bool AlignMiddle() {
            var changed = false;
            if(SelectionStatus == GISelectionStatus.MultipleItems) {
                var b = 0f;
                foreach(var gi in _selecteditems) b += gi.Topmm + gi.Heightmm / 2;
                b /= _selecteditems.Count;
                foreach(var gi in _selecteditems) {
                    var posmm = gi.Positionmm;
                    posmm.Y = b - gi.Heightmm / 2;
                    changed |= gi.SetPosition(posmm);
                }
            } else {
                var posmm = _selecteditems[0].Positionmm;
                posmm.Y = _card.Sizemm.Height / 2 - _selecteditems[0].Heightmm / 2;
                changed = _selecteditems[0].SetPosition(posmm);
            }
            return changed;
        }
        private bool AlignDown() {
            var changed = false;
            if(SelectionStatus == GISelectionStatus.MultipleItems) {
                var b = float.NegativeInfinity;
                foreach(var gi in _selecteditems) {
                    var btm = gi.Topmm + gi.Heightmm;
                    if(b < btm) b = btm;
                }
                foreach(var gi in _selecteditems) {
                    var posmm = gi.Positionmm;
                    posmm.Y = b - gi.Heightmm;
                    changed |= gi.SetPosition(posmm);
                }
            } else {
                var posmm = _selecteditems[0].Positionmm;
                posmm.Y = _card.Sizemm.Height - _selecteditems[0].Heightmm;
                changed = _selecteditems[0].SetPosition(posmm);
            }
            return changed;
        }

        public void AlignEqualGap() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus != GISelectionStatus.MultipleItems)
                throw new Exception("There are not items to be aligned");
            if(_selecteditems.Count == 2) return;

            VerticalSort(_selecteditems);
            var n = _selecteditems.Count;
            var htotal = _selecteditems[n - 1].Topmm + _selecteditems[n - 1].Heightmm
                       - _selecteditems[0].Topmm;
            var hsum = 0f;
            foreach(var gi in _selecteditems) hsum += gi.Heightmm;
            var gap = (htotal - hsum) / (n - 1);

            PushToUndoList();
            var changed = false;
            for(var i = 1; i < _selecteditems.Count; i++) {
                var posmm = _selecteditems[i].Positionmm;
                posmm.Y = _selecteditems[i - 1].Topmm + _selecteditems[i - 1].Heightmm + gap;
                changed |= _selecteditems[i].SetPosition(posmm);
            }
            if(!changed) PopFromUndoList();
            _drawsurface.Invalidate();
        }
        public void AlignNoGap() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus != GISelectionStatus.MultipleItems)
                throw new Exception("There are not items to be aligned");
            VerticalSort(_selecteditems);
            PushToUndoList();
            var changed = false;
            for(var i = 1; i < _selecteditems.Count; i++) {
                var posmm = _selecteditems[i].Positionmm;
                posmm.Y = _selecteditems[i - 1].Topmm + _selecteditems[i - 1].Heightmm;
                changed |= _selecteditems[i].SetPosition(posmm);
            }
            if(!changed) PopFromUndoList();
            _drawsurface.Invalidate();
        }
        public void AlignFontInterval() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(SelectionStatus != GISelectionStatus.MultipleItems)
                throw new Exception("There are not items to be aligned");
            VerticalSort(_selecteditems);
            PushToUndoList();
            var changed = false;
            for(var i = 1; i < _selecteditems.Count; i++) {
                var posmm = _selecteditems[i].Positionmm;
                posmm.Y = _selecteditems[i - 1].Topmm;
                if(_selecteditems[i - 1] is Characters)
                    posmm.Y += ((Characters)_selecteditems[i - 1]).FontHeightmm;
                else
                    posmm.Y += _selecteditems[i - 1].Heightmm;
                changed |= _selecteditems[i].SetPosition(posmm);
            }
            if(!changed) PopFromUndoList();
            _drawsurface.Invalidate();
        }
        private void VerticalSort(List<GraphicItem> list) {
            for(var i = 0; i < list.Count - 1; i++) {
                for(var j = i + 1; j < list.Count; j++)
                    if(list[j].Topmm < list[i].Topmm) {
                        var gi = list[i];
                        list[i] = list[j];
                        list[j] = gi;
                    }
            }
        }

        public void BeDrawnOn(Graphics g) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            _card.BeDrawnOn(g);
            using(var p = new Pen(NegativeColor, 0)) {
                foreach(var gi in _selecteditems)
                    g.DrawRectangle(p, gi.Leftmm, gi.Topmm, gi.Widthmm, gi.Heightmm);
            }
        }
        public void BePrintedOn(Graphics g) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            _card.BeDrawnOn(g);
        }

        public FileInfo File {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _file;
            }
        }
        public SizeF Sizemm {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _card.Sizemm;
            }
            set {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                PushToUndoList();
                if(_card.SetSize(value))
                    _drawsurface.Invalidate();
                else
                    PopFromUndoList();
            }
        }
        public Color Color {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _card.Color;
            }
            set {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                PushToUndoList();
                if(_card.SetColor(value))
                    _drawsurface.Invalidate();
                else
                    PopFromUndoList();
            }
        }
        public Color NegativeColor {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return Color.FromArgb(255 - _card.Color.R,
                                      255 - _card.Color.G,
                                      255 - _card.Color.B);
            }
        }
        public int[] CustomColors {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _card.CustomColors;
            }
            set {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                _card.CustomColors = value;
            }
        }

        [Serializable]
        private class UndoPackage {
            private NameCard _card;
            private List<GraphicItem> _selecteditems;
            public UndoPackage(NameCard card, List<GraphicItem> selecteditems) {
                _card = card;
                _selecteditems = selecteditems;
            }
            public void Restore(ref NameCard card, ref List<GraphicItem> selecteditems) {
                foreach(var gi in selecteditems) gi.Dispose();
                if(_card != null) card.Dispose();
                card = _card;
                selecteditems = _selecteditems;
            }
        }
    }
}
