﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Text;
using System.Windows.Forms;

namespace NameCard {
    [Serializable]
    class NameCard : IDisposable {
        //Essential members to be serialized
        private List<GraphicItem> _graphicitems = new List<GraphicItem>();
        private SizeF _sizemm;
        private Color _color = Color.White;
        private int[] _customcolors = null;

        //Constants
        [NonSerialized]
        private const string CLASS_NAME = "NameCard";

        //Working members
        [NonSerialized]
        private bool _disposed = false;

        public NameCard() {
            _sizemm = new SizeF(91f, 55f);
            _color = Color.White;
        }
        protected void Dispose(bool disposing) {
            if(disposing) {
                //Free manage resource
                foreach(var gi in _graphicitems) gi.Dispose();
            }
            //Free unmanage resource
            _disposed = true;
        }
        #region IDisposable Member
        public void Dispose() {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion
        ~NameCard() {
            Dispose(false);
        }

        public bool SetSize(SizeF sizemm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(_sizemm == sizemm) return false;
            if(sizemm == SizeF.Empty)
                throw new ArgumentException("The size is empty.");
            if(sizemm.Width <= 0)
                throw new ArgumentException("The width is negative or zero.");
            if(sizemm.Width == float.PositiveInfinity)
                throw new ArgumentException("The width is infinite.");
            if(sizemm.Height <= 0)
                throw new ArgumentException("The height is negative or zero.");
            if(sizemm.Height == float.PositiveInfinity)
                throw new ArgumentException("The height is infinite.");
            _sizemm = sizemm;
            return true;
        }
        public bool SetColor(Color color) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(_color == color) return false;
            if(color == Color.Empty)
                throw new ArgumentException("The color is empty.");
            _color = color;
            return true;
        }

        public void Add(GraphicItem gi) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(gi == null) throw new ArgumentNullException("gi");
            if(_graphicitems.IndexOf(gi) >= 0)
                throw new ArgumentException("The graphic item has been already included.");
            _graphicitems.Add(gi);
        }
        public void Remove(GraphicItem gi) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(gi == null) throw new ArgumentNullException("gi");
            if(!_graphicitems.Remove(gi))
                throw new ArgumentException("The graphic item has not been included yet.");
        }
        public int Count {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _graphicitems.Count;
            }
        }
        public GraphicItem this[int i] {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _graphicitems[i];
            }
        }
        public IEnumerator<GraphicItem> GetEnumerator() {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            foreach(var item in _graphicitems) yield return item;
        }
        public GraphicItem ItemHit(PointF posmm) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            foreach(GraphicItem gi in _graphicitems) if(gi.Hit(posmm)) return gi;
            return null;
        }

        public void BeDrawnOn(Graphics g) {
            if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
            if(_sizemm.IsEmpty)
                throw new ArgumentException("Sizemm is missing in NameCard.");
            if(_color.IsEmpty)
                throw new ArgumentException("Color is missing in NameCard.");
            using(var b = new SolidBrush(_color))
                g.FillRectangle(b, 0f, 0f, _sizemm.Width, _sizemm.Height);
            foreach(GraphicItem gi in _graphicitems) gi.BeDrawnOn(g);
        }

        public SizeF Sizemm {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _sizemm;
            }
        }
        public Color Color {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _color;
            }
        }
        public int[] CustomColors {
            get {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                return _customcolors;
            }
            set {
                if(_disposed) throw new ObjectDisposedException(CLASS_NAME);
                _customcolors = value;
            }
        }
    }
    [Serializable]
    struct FontInfo {
        public FontInfo(string familyname, float emsizept, FontStyle fontstyle, Color color) {
            FamilyName = familyname;
            EmSizePoint = emsizept;
            Style = fontstyle;
            Color = color;
        }
        public string FamilyName;
        public float EmSizePoint;
        public FontStyle Style;
        public Color Color;
        public Font CreateFont() {
            if(FamilyName == null || FamilyName.Length == 0)
                throw new ArgumentException("FamilyName is missing in FontInfo.");
            return new Font(FamilyName, EmSizePoint, Style);
        }
        public static bool operator ==(FontInfo a, FontInfo b) {
            if(a.FamilyName != b.FamilyName) return false;
            if(a.EmSizePoint != b.EmSizePoint) return false;
            if(a.Style != b.Style) return false;
            if(a.Color != b.Color) return false;
            return true;
        }
        public static bool operator !=(FontInfo a, FontInfo b) {
            return !(a == b);
        }
    }
    [Serializable]
    abstract class GraphicItem : IDisposable {
        private const string CLASS_NAME = "GraphicItem";
        protected bool _disposed = false;
        protected PointF _pos;
        protected SizeF _size;

        protected abstract void Dispose(bool disposing);
        #region IDisposable Member
        public void Dispose() {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion
        ~GraphicItem() {
            Dispose(false);
        }

        public virtual void BeDrawnOn(Graphics g) {
            if(_disposed)
                throw new System.ObjectDisposedException(CLASS_NAME);
            if(_size.IsEmpty)
                throw new ArgumentException("Sizemm is missing in GraphicItem.");
        }
        public bool Hit(PointF posmm) {
            if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);

            if(posmm.X < Leftmm || Leftmm + Widthmm < posmm.X) return false;
            if(posmm.Y < Topmm || Topmm + Heightmm < posmm.Y) return false;
            return true;
        }
        public virtual bool SetPosition(PointF posmm) {
            if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
            if(_pos == posmm) return false;
            _pos = posmm;
            return true;
        }

        public PointF Positionmm {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _pos;
            }
        }
        public float Topmm {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _pos.Y;
            }
        }
        public float Leftmm {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _pos.X;
            }
        }
        public SizeF Sizemm {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _size;
            }
        }
        public float Widthmm {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _size.Width;
            }
        }
        public float Heightmm {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _size.Height;
            }
        }
    }
    [Serializable]
    class Characters : GraphicItem {
        private const string CLASS_NAME = "Characters";
        private string _text;
        private FontInfo _font;
        [NonSerialized]
        private float _fontHeightmm;

        public Characters(string text, FontInfo font, PointF posmm, SizeF sizemm) {
            SetContents(text, font, posmm, sizemm);
        }
        public bool SetContents(string text, FontInfo font, PointF posmm, SizeF sizemm) {
            if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
            var changed = false;
            if(_text != text) { _text = text; changed = true; }
            if(_font != font) { _font = font; changed = true; }
            if(_pos != posmm) { _pos = posmm; changed = true; }
            if(_size != sizemm) { _size = sizemm; changed = true; }
            return changed;
        }
        protected override void Dispose(bool disposing) {
            if(disposing) {
                //Free manage resource
            }
            //Free unmanage resource
            _disposed = true;
        }

        public override void BeDrawnOn(Graphics g) {
            base.BeDrawnOn(g);
            if(_font.Color.IsEmpty)
                throw new ArgumentException("Color of FontInfo is missing in FontInfo.");
            if(_text == null)
                throw new ArgumentException("Text is missing.");
            using(Font f = _font.CreateFont())
            using(var sf = StringFormat.GenericTypographic)
            using(SolidBrush b = new SolidBrush(_font.Color)) {
                _fontHeightmm = f.GetHeight(g);
                g.TextRenderingHint = TextRenderingHint.AntiAlias;
                g.DrawString(_text, f, b, _pos, sf);
            }
        }

        public string Text {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _text;
            }
        }
        public FontInfo Font {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _font;
            }
        }
        public float FontHeightmm {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _fontHeightmm;
            }
        }
    }
    [Serializable]
    class BitmapImage : GraphicItem {
        private const string CLASS_NAME = "BitmapImage";
        private Bitmap _image;

        public BitmapImage(Bitmap image, PointF posmm, SizeF sizemm) {
            _image = image;
            _size = sizemm;
            _pos = posmm;
        }
        public bool SetSize(SizeF sizemm) {
            if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
            if(_size == sizemm) return false;
            _size = sizemm;
            return true;
        }
        protected override void Dispose(bool disposing) {
            if(disposing) {
                _image.Dispose();
            }
            //Free unmanage resource
            _disposed = true;
        }

        public override void BeDrawnOn(Graphics g) {
            base.BeDrawnOn(g);
            g.DrawImage(_image, new RectangleF(_pos, _size));
        }

        public Bitmap Image {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return (Bitmap)_image.Clone();
            }
        }
    }
    [Serializable]
    class GroupedItem : GraphicItem {
        private const string CLASS_NAME = "GroupedItem";
        private List<GraphicItem> _gis = new List<GraphicItem>();

        public GroupedItem(List<GraphicItem> items) {
            var r = new RectangleF(items[0].Positionmm, SizeF.Empty); ;
            foreach(var gi in items) {
                _gis.Add(gi);
                r = RectangleF.Union(r, new RectangleF(gi.Positionmm, gi.Sizemm));
            }
            _pos = r.Location;
            _size = r.Size;
        }
        public override bool SetPosition(PointF posmm) {
            if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
            var dx = posmm.X - _pos.X;
            var dy = posmm.Y - _pos.Y;
            if(dx == 0f && dy == 0f) return false;
            _pos = posmm;
            foreach(var gi in _gis) {
                var pt = gi.Positionmm;
                pt.X += dx;
                pt.Y += dy;
                gi.SetPosition(pt);
            }
            return true;
        }
        public void Clear() {
            _gis.Clear();
            this.Dispose();
        }
        protected override void Dispose(bool disposing) {
            if(disposing) {
                foreach(var gi in _gis) gi.Dispose();
                _gis.Clear();
            }
            //Free unmanage resource
            _disposed = true;
        }

        public override void BeDrawnOn(Graphics g) {
            base.BeDrawnOn(g);
            if(_gis == null)
                throw new ArgumentException("GroupedItem is empty.");
            foreach(var gi in _gis) gi.BeDrawnOn(g);
        }

        public int Count {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _gis.Count;
            }
        }
        public GraphicItem this[int i] {
            get {
                if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
                return _gis[i];
            }
        }
        public IEnumerator<GraphicItem> GetEnumerator() {
            if(_disposed) throw new System.ObjectDisposedException(CLASS_NAME);
            foreach(var gi in _gis) yield return gi;
        }
    }
}
